from puya.awst import wtypes
from puya.awst_build.intrinsic_models import ArgMapping, FunctionOpMapping

ENUM_CLASSES = {
    "puyapy._gen.Base64": {
        "URLEncoding": "URLEncoding",
        "StdEncoding": "StdEncoding",
    },
    "puyapy._gen.Ecdsa": {
        "Secp256k1": "Secp256k1",
        "Secp256r1": "Secp256r1",
    },
    "puyapy._gen.VrfVerify": {
        "VrfAlgorand": "VrfAlgorand",
    },
}

STUB_TO_AST_MAPPER = {
    "puyapy._gen.addw": [
        FunctionOpMapping(
            op_code="addw",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.app_opted_in": [
        FunctionOpMapping(
            op_code="app_opted_in",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.arg": [
        FunctionOpMapping(
            op_code="args",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="arg",
            immediates=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="arg_0",
            immediates=[],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="arg_1",
            immediates=[],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="arg_2",
            immediates=[],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="arg_3",
            immediates=[],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.balance": [
        FunctionOpMapping(
            op_code="balance",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.base64_decode": [
        FunctionOpMapping(
            op_code="base64_decode",
            immediates=[
                ArgMapping(
                    arg_name="e",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.bitlen": [
        FunctionOpMapping(
            op_code="bitlen",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.bsqrt": [
        FunctionOpMapping(
            op_code="bsqrt",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.biguint_wtype,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.biguint_wtype,
            ],
        ),
    ],
    "puyapy._gen.btoi": [
        FunctionOpMapping(
            op_code="btoi",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.bzero": [
        FunctionOpMapping(
            op_code="bzero",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.concat": [
        FunctionOpMapping(
            op_code="concat",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.divmodw": [
        FunctionOpMapping(
            op_code="divmodw",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="d",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.uint64_wtype,
                wtypes.uint64_wtype,
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.divw": [
        FunctionOpMapping(
            op_code="divw",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.ecdsa_pk_decompress": [
        FunctionOpMapping(
            op_code="ecdsa_pk_decompress",
            immediates=[
                ArgMapping(
                    arg_name="v",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.ecdsa_pk_recover": [
        FunctionOpMapping(
            op_code="ecdsa_pk_recover",
            immediates=[
                ArgMapping(
                    arg_name="v",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="d",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.ecdsa_verify": [
        FunctionOpMapping(
            op_code="ecdsa_verify",
            immediates=[
                ArgMapping(
                    arg_name="v",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="d",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="e",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.ed25519verify": [
        FunctionOpMapping(
            op_code="ed25519verify",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.ed25519verify_bare": [
        FunctionOpMapping(
            op_code="ed25519verify_bare",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.err": [
        FunctionOpMapping(
            op_code="err",
            immediates=[],
            stack_inputs=[],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.exit": [
        FunctionOpMapping(
            op_code="return",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.exp": [
        FunctionOpMapping(
            op_code="exp",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.expw": [
        FunctionOpMapping(
            op_code="expw",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.extract": [
        FunctionOpMapping(
            op_code="extract3",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="extract",
            immediates=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.extract_uint16": [
        FunctionOpMapping(
            op_code="extract_uint16",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.extract_uint32": [
        FunctionOpMapping(
            op_code="extract_uint32",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.extract_uint64": [
        FunctionOpMapping(
            op_code="extract_uint64",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.gaid": [
        FunctionOpMapping(
            op_code="gaids",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gaid",
            immediates=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.getbit": [
        FunctionOpMapping(
            op_code="getbit",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.getbyte": [
        FunctionOpMapping(
            op_code="getbyte",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.gload_bytes": [
        FunctionOpMapping(
            op_code="gloadss",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gload",
            immediates=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gloads",
            immediates=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.gload_uint64": [
        FunctionOpMapping(
            op_code="gloadss",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gload",
            immediates=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gloads",
            immediates=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.itob": [
        FunctionOpMapping(
            op_code="itob",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.keccak256": [
        FunctionOpMapping(
            op_code="keccak256",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.log": [
        FunctionOpMapping(
            op_code="log",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.min_balance": [
        FunctionOpMapping(
            op_code="min_balance",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.mulw": [
        FunctionOpMapping(
            op_code="mulw",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.replace": [
        FunctionOpMapping(
            op_code="replace3",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="replace2",
            immediates=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.setbit_bytes": [
        FunctionOpMapping(
            op_code="setbit",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.setbit_uint64": [
        FunctionOpMapping(
            op_code="setbit",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.setbyte": [
        FunctionOpMapping(
            op_code="setbyte",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.sha256": [
        FunctionOpMapping(
            op_code="sha256",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.sha3_256": [
        FunctionOpMapping(
            op_code="sha3_256",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.sha512_256": [
        FunctionOpMapping(
            op_code="sha512_256",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.shl": [
        FunctionOpMapping(
            op_code="shl",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.shr": [
        FunctionOpMapping(
            op_code="shr",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.sqrt": [
        FunctionOpMapping(
            op_code="sqrt",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.substring": [
        FunctionOpMapping(
            op_code="substring3",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="substring",
            immediates=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.vrf_verify": [
        FunctionOpMapping(
            op_code="vrf_verify",
            immediates=[
                ArgMapping(
                    arg_name="s",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AcctParamsGet.acct_balance": [
        FunctionOpMapping(
            op_code="acct_params_get",
            immediates=[
                "AcctBalance",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AcctParamsGet.acct_min_balance": [
        FunctionOpMapping(
            op_code="acct_params_get",
            immediates=[
                "AcctMinBalance",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AcctParamsGet.acct_auth_addr": [
        FunctionOpMapping(
            op_code="acct_params_get",
            immediates=[
                "AcctAuthAddr",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AcctParamsGet.acct_total_num_uint": [
        FunctionOpMapping(
            op_code="acct_params_get",
            immediates=[
                "AcctTotalNumUint",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AcctParamsGet.acct_total_num_byte_slice": [
        FunctionOpMapping(
            op_code="acct_params_get",
            immediates=[
                "AcctTotalNumByteSlice",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AcctParamsGet.acct_total_extra_app_pages": [
        FunctionOpMapping(
            op_code="acct_params_get",
            immediates=[
                "AcctTotalExtraAppPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AcctParamsGet.acct_total_apps_created": [
        FunctionOpMapping(
            op_code="acct_params_get",
            immediates=[
                "AcctTotalAppsCreated",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AcctParamsGet.acct_total_apps_opted_in": [
        FunctionOpMapping(
            op_code="acct_params_get",
            immediates=[
                "AcctTotalAppsOptedIn",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AcctParamsGet.acct_total_assets_created": [
        FunctionOpMapping(
            op_code="acct_params_get",
            immediates=[
                "AcctTotalAssetsCreated",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AcctParamsGet.acct_total_assets": [
        FunctionOpMapping(
            op_code="acct_params_get",
            immediates=[
                "AcctTotalAssets",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AcctParamsGet.acct_total_boxes": [
        FunctionOpMapping(
            op_code="acct_params_get",
            immediates=[
                "AcctTotalBoxes",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AcctParamsGet.acct_total_box_bytes": [
        FunctionOpMapping(
            op_code="acct_params_get",
            immediates=[
                "AcctTotalBoxBytes",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppGlobals.get_bytes": [
        FunctionOpMapping(
            op_code="app_global_get",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppGlobals.get_uint64": [
        FunctionOpMapping(
            op_code="app_global_get",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppGlobals.get_ex_bytes": [
        FunctionOpMapping(
            op_code="app_global_get_ex",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppGlobals.get_ex_uint64": [
        FunctionOpMapping(
            op_code="app_global_get_ex",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppGlobals.delete": [
        FunctionOpMapping(
            op_code="app_global_del",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.AppGlobals.put": [
        FunctionOpMapping(
            op_code="app_global_put",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.AppLocals.get_bytes": [
        FunctionOpMapping(
            op_code="app_local_get",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppLocals.get_uint64": [
        FunctionOpMapping(
            op_code="app_local_get",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppLocals.get_ex_bytes": [
        FunctionOpMapping(
            op_code="app_local_get_ex",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppLocals.get_ex_uint64": [
        FunctionOpMapping(
            op_code="app_local_get_ex",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppLocals.delete": [
        FunctionOpMapping(
            op_code="app_local_del",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.AppLocals.put": [
        FunctionOpMapping(
            op_code="app_local_put",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.AppParamsGet.app_approval_program": [
        FunctionOpMapping(
            op_code="app_params_get",
            immediates=[
                "AppApprovalProgram",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppParamsGet.app_clear_state_program": [
        FunctionOpMapping(
            op_code="app_params_get",
            immediates=[
                "AppClearStateProgram",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppParamsGet.app_global_num_uint": [
        FunctionOpMapping(
            op_code="app_params_get",
            immediates=[
                "AppGlobalNumUint",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppParamsGet.app_global_num_byte_slice": [
        FunctionOpMapping(
            op_code="app_params_get",
            immediates=[
                "AppGlobalNumByteSlice",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppParamsGet.app_local_num_uint": [
        FunctionOpMapping(
            op_code="app_params_get",
            immediates=[
                "AppLocalNumUint",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppParamsGet.app_local_num_byte_slice": [
        FunctionOpMapping(
            op_code="app_params_get",
            immediates=[
                "AppLocalNumByteSlice",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppParamsGet.app_extra_program_pages": [
        FunctionOpMapping(
            op_code="app_params_get",
            immediates=[
                "AppExtraProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppParamsGet.app_creator": [
        FunctionOpMapping(
            op_code="app_params_get",
            immediates=[
                "AppCreator",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AppParamsGet.app_address": [
        FunctionOpMapping(
            op_code="app_params_get",
            immediates=[
                "AppAddress",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AssetHoldingGet.asset_balance": [
        FunctionOpMapping(
            op_code="asset_holding_get",
            immediates=[
                "AssetBalance",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AssetHoldingGet.asset_frozen": [
        FunctionOpMapping(
            op_code="asset_holding_get",
            immediates=[
                "AssetFrozen",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bool_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AssetParamsGet.asset_total": [
        FunctionOpMapping(
            op_code="asset_params_get",
            immediates=[
                "AssetTotal",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AssetParamsGet.asset_decimals": [
        FunctionOpMapping(
            op_code="asset_params_get",
            immediates=[
                "AssetDecimals",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AssetParamsGet.asset_default_frozen": [
        FunctionOpMapping(
            op_code="asset_params_get",
            immediates=[
                "AssetDefaultFrozen",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bool_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AssetParamsGet.asset_unit_name": [
        FunctionOpMapping(
            op_code="asset_params_get",
            immediates=[
                "AssetUnitName",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AssetParamsGet.asset_name": [
        FunctionOpMapping(
            op_code="asset_params_get",
            immediates=[
                "AssetName",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AssetParamsGet.asset_url": [
        FunctionOpMapping(
            op_code="asset_params_get",
            immediates=[
                "AssetURL",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AssetParamsGet.asset_metadata_hash": [
        FunctionOpMapping(
            op_code="asset_params_get",
            immediates=[
                "AssetMetadataHash",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AssetParamsGet.asset_manager": [
        FunctionOpMapping(
            op_code="asset_params_get",
            immediates=[
                "AssetManager",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AssetParamsGet.asset_reserve": [
        FunctionOpMapping(
            op_code="asset_params_get",
            immediates=[
                "AssetReserve",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AssetParamsGet.asset_freeze": [
        FunctionOpMapping(
            op_code="asset_params_get",
            immediates=[
                "AssetFreeze",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AssetParamsGet.asset_clawback": [
        FunctionOpMapping(
            op_code="asset_params_get",
            immediates=[
                "AssetClawback",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.AssetParamsGet.asset_creator": [
        FunctionOpMapping(
            op_code="asset_params_get",
            immediates=[
                "AssetCreator",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.Block.blk_seed": [
        FunctionOpMapping(
            op_code="block",
            immediates=[
                "BlkSeed",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Block.blk_timestamp": [
        FunctionOpMapping(
            op_code="block",
            immediates=[
                "BlkTimestamp",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Box.create": [
        FunctionOpMapping(
            op_code="box_create",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.Box.delete": [
        FunctionOpMapping(
            op_code="box_del",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.Box.extract": [
        FunctionOpMapping(
            op_code="box_extract",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Box.get": [
        FunctionOpMapping(
            op_code="box_get",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.Box.length": [
        FunctionOpMapping(
            op_code="box_len",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.Box.put": [
        FunctionOpMapping(
            op_code="box_put",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.Box.replace": [
        FunctionOpMapping(
            op_code="box_replace",
            immediates=[],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="c",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.begin": [
        FunctionOpMapping(
            op_code="itxn_begin",
            immediates=[],
            stack_inputs=[],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.next": [
        FunctionOpMapping(
            op_code="itxn_next",
            immediates=[],
            stack_inputs=[],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.submit": [
        FunctionOpMapping(
            op_code="itxn_submit",
            immediates=[],
            stack_inputs=[],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_sender": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "Sender",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_fee": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "Fee",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_note": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "Note",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_receiver": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "Receiver",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_amount": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "Amount",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_close_remainder_to": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "CloseRemainderTo",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_vote_pk": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "VotePK",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_selection_pk": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "SelectionPK",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_vote_first": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "VoteFirst",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_vote_last": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "VoteLast",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_vote_key_dilution": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "VoteKeyDilution",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_type": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "Type",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_type_enum": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "TypeEnum",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_xfer_asset": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "XferAsset",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_asset_amount": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "AssetAmount",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_asset_sender": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "AssetSender",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_asset_receiver": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "AssetReceiver",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_asset_close_to": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "AssetCloseTo",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_application_id": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ApplicationID",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_on_completion": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "OnCompletion",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_application_args": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ApplicationArgs",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_accounts": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "Accounts",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_approval_program": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ApprovalProgram",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_clear_state_program": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ClearStateProgram",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_rekey_to": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "RekeyTo",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_config_asset": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ConfigAsset",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_config_asset_total": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ConfigAssetTotal",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_config_asset_decimals": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ConfigAssetDecimals",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_config_asset_default_frozen": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ConfigAssetDefaultFrozen",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bool_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_config_asset_unit_name": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ConfigAssetUnitName",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_config_asset_name": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ConfigAssetName",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_config_asset_url": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ConfigAssetURL",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_config_asset_metadata_hash": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ConfigAssetMetadataHash",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_config_asset_manager": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ConfigAssetManager",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_config_asset_reserve": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ConfigAssetReserve",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_config_asset_freeze": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ConfigAssetFreeze",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_config_asset_clawback": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ConfigAssetClawback",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_freeze_asset": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "FreezeAsset",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_freeze_asset_account": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "FreezeAssetAccount",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.account_wtype,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_freeze_asset_frozen": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "FreezeAssetFrozen",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bool_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_assets": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "Assets",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_applications": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "Applications",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_global_num_uint": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "GlobalNumUint",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_global_num_byte_slice": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "GlobalNumByteSlice",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_local_num_uint": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "LocalNumUint",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_local_num_byte_slice": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "LocalNumByteSlice",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_extra_program_pages": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ExtraProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_nonparticipation": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "Nonparticipation",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bool_wtype,
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_state_proof_pk": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "StateProofPK",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_approval_program_pages": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ApprovalProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.CreateInnerTransaction.set_clear_state_program_pages": [
        FunctionOpMapping(
            op_code="itxn_field",
            immediates=[
                "ClearStateProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[],
        ),
    ],
    "puyapy._gen.Global.min_txn_fee": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "MinTxnFee",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Global.min_balance": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "MinBalance",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Global.max_txn_life": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "MaxTxnLife",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Global.zero_address": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "ZeroAddress",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Global.group_size": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "GroupSize",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Global.logic_sig_version": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "LogicSigVersion",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Global.round": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "Round",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Global.latest_timestamp": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "LatestTimestamp",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Global.current_application_id": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "CurrentApplicationID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Global.creator_address": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "CreatorAddress",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Global.current_application_address": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "CurrentApplicationAddress",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Global.group_id": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "GroupID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Global.opcode_budget": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "OpcodeBudget",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Global.caller_application_id": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "CallerApplicationID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Global.caller_application_address": [
        FunctionOpMapping(
            op_code="global",
            immediates=[
                "CallerApplicationAddress",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.sender": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "Sender",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.fee": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "Fee",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.first_valid": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "FirstValid",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.first_valid_time": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "FirstValidTime",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.last_valid": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "LastValid",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.note": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "Note",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.lease": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "Lease",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.receiver": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "Receiver",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.amount": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "Amount",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.close_remainder_to": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "CloseRemainderTo",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.vote_pk": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "VotePK",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.selection_pk": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "SelectionPK",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.vote_first": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "VoteFirst",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.vote_last": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "VoteLast",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.vote_key_dilution": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "VoteKeyDilution",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.type": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "Type",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.type_enum": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "TypeEnum",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.xfer_asset": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "XferAsset",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.asset_amount": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "AssetAmount",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.asset_sender": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "AssetSender",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.asset_receiver": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "AssetReceiver",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.asset_close_to": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "AssetCloseTo",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.group_index": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "GroupIndex",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.tx_id": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "TxID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.application_id": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ApplicationID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.on_completion": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "OnCompletion",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.application_args": [
        FunctionOpMapping(
            op_code="itxnas",
            immediates=[
                "ApplicationArgs",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="itxna",
            immediates=[
                "ApplicationArgs",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.num_app_args": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "NumAppArgs",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.accounts": [
        FunctionOpMapping(
            op_code="itxnas",
            immediates=[
                "Accounts",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="itxna",
            immediates=[
                "Accounts",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.num_accounts": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "NumAccounts",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.approval_program": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ApprovalProgram",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.clear_state_program": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ClearStateProgram",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.rekey_to": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "RekeyTo",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.config_asset": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ConfigAsset",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.config_asset_total": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ConfigAssetTotal",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.config_asset_decimals": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ConfigAssetDecimals",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.config_asset_default_frozen": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ConfigAssetDefaultFrozen",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.config_asset_unit_name": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ConfigAssetUnitName",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.config_asset_name": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ConfigAssetName",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.config_asset_url": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ConfigAssetURL",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.config_asset_metadata_hash": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ConfigAssetMetadataHash",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.config_asset_manager": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ConfigAssetManager",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.config_asset_reserve": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ConfigAssetReserve",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.config_asset_freeze": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ConfigAssetFreeze",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.config_asset_clawback": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ConfigAssetClawback",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.freeze_asset": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "FreezeAsset",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.freeze_asset_account": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "FreezeAssetAccount",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.freeze_asset_frozen": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "FreezeAssetFrozen",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.assets": [
        FunctionOpMapping(
            op_code="itxnas",
            immediates=[
                "Assets",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="itxna",
            immediates=[
                "Assets",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.num_assets": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "NumAssets",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.applications": [
        FunctionOpMapping(
            op_code="itxnas",
            immediates=[
                "Applications",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="itxna",
            immediates=[
                "Applications",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.num_applications": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "NumApplications",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.global_num_uint": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "GlobalNumUint",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.global_num_byte_slice": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "GlobalNumByteSlice",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.local_num_uint": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "LocalNumUint",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.local_num_byte_slice": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "LocalNumByteSlice",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.extra_program_pages": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "ExtraProgramPages",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.nonparticipation": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "Nonparticipation",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.logs": [
        FunctionOpMapping(
            op_code="itxnas",
            immediates=[
                "Logs",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="itxna",
            immediates=[
                "Logs",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.num_logs": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "NumLogs",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.created_asset_id": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "CreatedAssetID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.created_application_id": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "CreatedApplicationID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.last_log": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "LastLog",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.state_proof_pk": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "StateProofPK",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.approval_program_pages": [
        FunctionOpMapping(
            op_code="itxnas",
            immediates=[
                "ApprovalProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="itxna",
            immediates=[
                "ApprovalProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.num_approval_program_pages": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "NumApprovalProgramPages",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.clear_state_program_pages": [
        FunctionOpMapping(
            op_code="itxnas",
            immediates=[
                "ClearStateProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="itxna",
            immediates=[
                "ClearStateProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransaction.num_clear_state_program_pages": [
        FunctionOpMapping(
            op_code="itxn",
            immediates=[
                "NumClearStateProgramPages",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.sender": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Sender",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.fee": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Fee",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.first_valid": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "FirstValid",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.first_valid_time": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "FirstValidTime",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.last_valid": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "LastValid",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.note": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Note",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.lease": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Lease",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.receiver": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Receiver",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.amount": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Amount",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.close_remainder_to": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "CloseRemainderTo",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.vote_pk": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "VotePK",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.selection_pk": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "SelectionPK",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.vote_first": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "VoteFirst",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.vote_last": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "VoteLast",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.vote_key_dilution": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "VoteKeyDilution",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.type": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Type",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.type_enum": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "TypeEnum",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.xfer_asset": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "XferAsset",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.asset_amount": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "AssetAmount",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.asset_sender": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "AssetSender",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.asset_receiver": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "AssetReceiver",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.asset_close_to": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "AssetCloseTo",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.group_index": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "GroupIndex",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.tx_id": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "TxID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.application_id": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ApplicationID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.on_completion": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "OnCompletion",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.application_args": [
        FunctionOpMapping(
            op_code="gitxnas",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ApplicationArgs",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gitxna",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ApplicationArgs",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.num_app_args": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "NumAppArgs",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.accounts": [
        FunctionOpMapping(
            op_code="gitxnas",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Accounts",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gitxna",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Accounts",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.num_accounts": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "NumAccounts",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.approval_program": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ApprovalProgram",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.clear_state_program": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ClearStateProgram",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.rekey_to": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "RekeyTo",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.config_asset": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ConfigAsset",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.config_asset_total": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ConfigAssetTotal",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.config_asset_decimals": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ConfigAssetDecimals",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.config_asset_default_frozen": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ConfigAssetDefaultFrozen",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.config_asset_unit_name": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ConfigAssetUnitName",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.config_asset_name": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ConfigAssetName",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.config_asset_url": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ConfigAssetURL",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.config_asset_metadata_hash": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ConfigAssetMetadataHash",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.config_asset_manager": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ConfigAssetManager",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.config_asset_reserve": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ConfigAssetReserve",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.config_asset_freeze": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ConfigAssetFreeze",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.config_asset_clawback": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ConfigAssetClawback",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.freeze_asset": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "FreezeAsset",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.freeze_asset_account": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "FreezeAssetAccount",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.freeze_asset_frozen": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "FreezeAssetFrozen",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.assets": [
        FunctionOpMapping(
            op_code="gitxnas",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Assets",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gitxna",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Assets",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.num_assets": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "NumAssets",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.applications": [
        FunctionOpMapping(
            op_code="gitxnas",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Applications",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gitxna",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Applications",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.num_applications": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "NumApplications",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.global_num_uint": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "GlobalNumUint",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.global_num_byte_slice": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "GlobalNumByteSlice",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.local_num_uint": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "LocalNumUint",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.local_num_byte_slice": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "LocalNumByteSlice",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.extra_program_pages": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ExtraProgramPages",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.nonparticipation": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Nonparticipation",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.logs": [
        FunctionOpMapping(
            op_code="gitxnas",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Logs",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gitxna",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "Logs",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.num_logs": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "NumLogs",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.created_asset_id": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "CreatedAssetID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.created_application_id": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "CreatedApplicationID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.last_log": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "LastLog",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.state_proof_pk": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "StateProofPK",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.approval_program_pages": [
        FunctionOpMapping(
            op_code="gitxnas",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ApprovalProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gitxna",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ApprovalProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.num_approval_program_pages": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "NumApprovalProgramPages",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.clear_state_program_pages": [
        FunctionOpMapping(
            op_code="gitxnas",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ClearStateProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gitxna",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "ClearStateProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.InnerTransactionGroup.num_clear_state_program_pages": [
        FunctionOpMapping(
            op_code="gitxn",
            immediates=[
                ArgMapping(
                    arg_name="t",
                    allowed_types=[
                        int,
                    ],
                ),
                "NumClearStateProgramPages",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.JsonRef.json_string": [
        FunctionOpMapping(
            op_code="json_ref",
            immediates=[
                "JSONString",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.JsonRef.json_uint64": [
        FunctionOpMapping(
            op_code="json_ref",
            immediates=[
                "JSONUint64",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.JsonRef.json_object": [
        FunctionOpMapping(
            op_code="json_ref",
            immediates=[
                "JSONObject",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.bytes_wtype,
                        bytes,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.sender": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "Sender",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.fee": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "Fee",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.first_valid": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "FirstValid",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.first_valid_time": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "FirstValidTime",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.last_valid": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "LastValid",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.note": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "Note",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.lease": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "Lease",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.receiver": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "Receiver",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.amount": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "Amount",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.close_remainder_to": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "CloseRemainderTo",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.vote_pk": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "VotePK",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.selection_pk": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "SelectionPK",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.vote_first": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "VoteFirst",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.vote_last": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "VoteLast",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.vote_key_dilution": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "VoteKeyDilution",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.type": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "Type",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.type_enum": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "TypeEnum",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.xfer_asset": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "XferAsset",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.asset_amount": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "AssetAmount",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.asset_sender": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "AssetSender",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.asset_receiver": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "AssetReceiver",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.asset_close_to": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "AssetCloseTo",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.group_index": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "GroupIndex",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.tx_id": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "TxID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.application_id": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ApplicationID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.on_completion": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "OnCompletion",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.application_args": [
        FunctionOpMapping(
            op_code="txnas",
            immediates=[
                "ApplicationArgs",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="txna",
            immediates=[
                "ApplicationArgs",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.num_app_args": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "NumAppArgs",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.accounts": [
        FunctionOpMapping(
            op_code="txnas",
            immediates=[
                "Accounts",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="txna",
            immediates=[
                "Accounts",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.num_accounts": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "NumAccounts",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.approval_program": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ApprovalProgram",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.clear_state_program": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ClearStateProgram",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.rekey_to": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "RekeyTo",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.config_asset": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ConfigAsset",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.config_asset_total": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ConfigAssetTotal",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.config_asset_decimals": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ConfigAssetDecimals",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.config_asset_default_frozen": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ConfigAssetDefaultFrozen",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.config_asset_unit_name": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ConfigAssetUnitName",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.config_asset_name": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ConfigAssetName",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.config_asset_url": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ConfigAssetURL",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.config_asset_metadata_hash": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ConfigAssetMetadataHash",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.config_asset_manager": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ConfigAssetManager",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.config_asset_reserve": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ConfigAssetReserve",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.config_asset_freeze": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ConfigAssetFreeze",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.config_asset_clawback": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ConfigAssetClawback",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.freeze_asset": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "FreezeAsset",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.freeze_asset_account": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "FreezeAssetAccount",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.freeze_asset_frozen": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "FreezeAssetFrozen",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.assets": [
        FunctionOpMapping(
            op_code="txnas",
            immediates=[
                "Assets",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="txna",
            immediates=[
                "Assets",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.num_assets": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "NumAssets",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.applications": [
        FunctionOpMapping(
            op_code="txnas",
            immediates=[
                "Applications",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="txna",
            immediates=[
                "Applications",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.num_applications": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "NumApplications",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.global_num_uint": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "GlobalNumUint",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.global_num_byte_slice": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "GlobalNumByteSlice",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.local_num_uint": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "LocalNumUint",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.local_num_byte_slice": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "LocalNumByteSlice",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.extra_program_pages": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "ExtraProgramPages",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.nonparticipation": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "Nonparticipation",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.logs": [
        FunctionOpMapping(
            op_code="txnas",
            immediates=[
                "Logs",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="txna",
            immediates=[
                "Logs",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.num_logs": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "NumLogs",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.created_asset_id": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "CreatedAssetID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.created_application_id": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "CreatedApplicationID",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.last_log": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "LastLog",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.state_proof_pk": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "StateProofPK",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.approval_program_pages": [
        FunctionOpMapping(
            op_code="txnas",
            immediates=[
                "ApprovalProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="txna",
            immediates=[
                "ApprovalProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.num_approval_program_pages": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "NumApprovalProgramPages",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.clear_state_program_pages": [
        FunctionOpMapping(
            op_code="txnas",
            immediates=[
                "ClearStateProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="txna",
            immediates=[
                "ClearStateProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.Transaction.num_clear_state_program_pages": [
        FunctionOpMapping(
            op_code="txn",
            immediates=[
                "NumClearStateProgramPages",
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.sender": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "Sender",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "Sender",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.fee": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "Fee",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "Fee",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.first_valid": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "FirstValid",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "FirstValid",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.first_valid_time": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "FirstValidTime",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "FirstValidTime",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.last_valid": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "LastValid",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "LastValid",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.note": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "Note",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "Note",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.lease": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "Lease",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "Lease",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.receiver": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "Receiver",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "Receiver",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.amount": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "Amount",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "Amount",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.close_remainder_to": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "CloseRemainderTo",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "CloseRemainderTo",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.vote_pk": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "VotePK",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "VotePK",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.selection_pk": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "SelectionPK",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "SelectionPK",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.vote_first": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "VoteFirst",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "VoteFirst",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.vote_last": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "VoteLast",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "VoteLast",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.vote_key_dilution": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "VoteKeyDilution",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "VoteKeyDilution",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.type": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "Type",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "Type",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.type_enum": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "TypeEnum",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "TypeEnum",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.xfer_asset": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "XferAsset",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "XferAsset",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.asset_amount": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "AssetAmount",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "AssetAmount",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.asset_sender": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "AssetSender",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "AssetSender",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.asset_receiver": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "AssetReceiver",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "AssetReceiver",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.asset_close_to": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "AssetCloseTo",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "AssetCloseTo",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.group_index": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "GroupIndex",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "GroupIndex",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.tx_id": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "TxID",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "TxID",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.application_id": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ApplicationID",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ApplicationID",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.on_completion": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "OnCompletion",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "OnCompletion",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.application_args": [
        FunctionOpMapping(
            op_code="gtxnsas",
            immediates=[
                "ApplicationArgs",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxnsa",
            immediates=[
                "ApplicationArgs",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxna",
            immediates=[
                "ApplicationArgs",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxnas",
            immediates=[
                "ApplicationArgs",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.num_app_args": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "NumAppArgs",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "NumAppArgs",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.accounts": [
        FunctionOpMapping(
            op_code="gtxnsas",
            immediates=[
                "Accounts",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxnsa",
            immediates=[
                "Accounts",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxna",
            immediates=[
                "Accounts",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxnas",
            immediates=[
                "Accounts",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.num_accounts": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "NumAccounts",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "NumAccounts",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.approval_program": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ApprovalProgram",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ApprovalProgram",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.clear_state_program": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ClearStateProgram",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ClearStateProgram",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.rekey_to": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "RekeyTo",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "RekeyTo",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.config_asset": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ConfigAsset",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ConfigAsset",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.config_asset_total": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ConfigAssetTotal",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ConfigAssetTotal",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.config_asset_decimals": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ConfigAssetDecimals",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ConfigAssetDecimals",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.config_asset_default_frozen": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ConfigAssetDefaultFrozen",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ConfigAssetDefaultFrozen",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.config_asset_unit_name": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ConfigAssetUnitName",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ConfigAssetUnitName",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.config_asset_name": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ConfigAssetName",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ConfigAssetName",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.config_asset_url": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ConfigAssetURL",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ConfigAssetURL",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.config_asset_metadata_hash": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ConfigAssetMetadataHash",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ConfigAssetMetadataHash",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.config_asset_manager": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ConfigAssetManager",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ConfigAssetManager",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.config_asset_reserve": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ConfigAssetReserve",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ConfigAssetReserve",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.config_asset_freeze": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ConfigAssetFreeze",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ConfigAssetFreeze",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.config_asset_clawback": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ConfigAssetClawback",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ConfigAssetClawback",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.freeze_asset": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "FreezeAsset",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "FreezeAsset",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.freeze_asset_account": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "FreezeAssetAccount",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "FreezeAssetAccount",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.account_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.freeze_asset_frozen": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "FreezeAssetFrozen",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "FreezeAssetFrozen",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.assets": [
        FunctionOpMapping(
            op_code="gtxnsas",
            immediates=[
                "Assets",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxnsa",
            immediates=[
                "Assets",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxna",
            immediates=[
                "Assets",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxnas",
            immediates=[
                "Assets",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.num_assets": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "NumAssets",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "NumAssets",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.applications": [
        FunctionOpMapping(
            op_code="gtxnsas",
            immediates=[
                "Applications",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxnsa",
            immediates=[
                "Applications",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxna",
            immediates=[
                "Applications",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxnas",
            immediates=[
                "Applications",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.num_applications": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "NumApplications",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "NumApplications",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.global_num_uint": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "GlobalNumUint",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "GlobalNumUint",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.global_num_byte_slice": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "GlobalNumByteSlice",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "GlobalNumByteSlice",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.local_num_uint": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "LocalNumUint",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "LocalNumUint",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.local_num_byte_slice": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "LocalNumByteSlice",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "LocalNumByteSlice",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.extra_program_pages": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "ExtraProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "ExtraProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.nonparticipation": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "Nonparticipation",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "Nonparticipation",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bool_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.logs": [
        FunctionOpMapping(
            op_code="gtxnsas",
            immediates=[
                "Logs",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxnsa",
            immediates=[
                "Logs",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxna",
            immediates=[
                "Logs",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxnas",
            immediates=[
                "Logs",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.num_logs": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "NumLogs",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "NumLogs",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.created_asset_id": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "CreatedAssetID",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "CreatedAssetID",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.created_application_id": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "CreatedApplicationID",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "CreatedApplicationID",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.last_log": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "LastLog",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "LastLog",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.state_proof_pk": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "StateProofPK",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "StateProofPK",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.approval_program_pages": [
        FunctionOpMapping(
            op_code="gtxnsas",
            immediates=[
                "ApprovalProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxnsa",
            immediates=[
                "ApprovalProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxna",
            immediates=[
                "ApprovalProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxnas",
            immediates=[
                "ApprovalProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.num_approval_program_pages": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "NumApprovalProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "NumApprovalProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.clear_state_program_pages": [
        FunctionOpMapping(
            op_code="gtxnsas",
            immediates=[
                "ClearStateProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxnsa",
            immediates=[
                "ClearStateProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxna",
            immediates=[
                "ClearStateProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        int,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxnas",
            immediates=[
                "ClearStateProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="b",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.bytes_wtype,
            ],
        ),
    ],
    "puyapy._gen.TransactionGroup.num_clear_state_program_pages": [
        FunctionOpMapping(
            op_code="gtxns",
            immediates=[
                "NumClearStateProgramPages",
            ],
            stack_inputs=[
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        wtypes.uint64_wtype,
                        int,
                    ],
                ),
            ],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
        FunctionOpMapping(
            op_code="gtxn",
            immediates=[
                "NumClearStateProgramPages",
                ArgMapping(
                    arg_name="a",
                    allowed_types=[
                        str,
                    ],
                ),
            ],
            stack_inputs=[],
            stack_outputs=[
                wtypes.uint64_wtype,
            ],
        ),
    ],
}
